/*
 * $Id:$
 *
 * Copyright (c) 2016, Juniper Networks, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example COPYING); If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"
#include "int64.h"
#include <stdio.h>
#include "drivethres.h"
#include "utility.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef _WIN32
#include <io.h> // access()
#endif

#include <stdexcept>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MODEL_LEN 41

const char * drivethres_cpp_cvsid =
		"$Id: drivethres.cpp 3460 2014-07-28 21:19:33Z chrfranke $"
                DRIVETHRES_H_CVSID;

const struct ata_vendor_attributes unigen_attr[] = {
    {   1 , "Raw Read Error Rate     "},
    {   9 , "Power On Hours Count    "},
    {  12 , "Power Cycle Count       "},
    { 184 , "Initial Bad Block Count "},
    { 195 , "Program Failure Block Ct"},
    { 196 , "Erase Failure Block Ct  "},
    { 197 , "Read Failure Block Ct   "},
    { 198 , "Total Read Sectors Ct   "},
    { 199 , "Total Write Sectors Ct  "},
    { 200 , "Total Read Commands Ct  "},
    { 201 , "Total Write Commands Ct "},
    { 202 , "Total Err Bits Flash Ct "},
    { 203 , "Rd Sect correc Bit err  "},
    { 204 , "Bad Block Full Flag     "},
    { 205 , "Max PE Count Spec       "},
    { 206 , "Erase Count Min         "},
    { 207 , "Erase Count Max         "},
    { 208 , "Erase Count Average     "},
    { 209 , "Remaining Life (%)      "},
    { 210 , "BBM Error Log           "},
    { 000 , "END OF VENDOR ATTRIBUTE "}
};

//Configuration file for the threshold limits
static const char * configfile;

class drive_thres_cfg {
public:
  drive_thres_cfg(const char *);

  unsigned size() const
  { return m_thres_cfg_tab.size(); }

  const char * get_drive_name()
  { return series_name; }

  const attr_thres_cfg & operator[](unsigned i)
  { return m_thres_cfg_tab[i]; }

  void push_back(const attr_thres_cfg &);
private:
  char series_name[MODEL_LEN+1];
  std::vector<attr_thres_cfg> m_thres_cfg_tab;

};

drive_thres_cfg::drive_thres_cfg(const char *vendor_name)
{
  strncpy(series_name, vendor_name, MODEL_LEN);
}


void drive_thres_cfg::push_back(const attr_thres_cfg & src_attr_cfg)
{
  attr_thres_cfg dest_attr_cfg = src_attr_cfg;
  m_thres_cfg_tab.push_back(dest_attr_cfg);
}

// Thershold data base
static std::vector<drive_thres_cfg> thres_db;

// The function converts the arg to int and checks the range
// return val :
// -1 if the arg is null or value is not in the range
// value if the arg is valid one
static int GetValue(const char *arg,  const char *token, int lineno,
		    const char *cfgfile, int min, int max, int base)
{
  // make sure argument is there
  if (!arg) {
    pout("File %s line %d : Directive: %s takes integer argument "
         "from %d to %d.\n",
	     cfgfile, lineno, token, min, max);
    return -1;
  }

  // get argument value, check that passed token is in-range
  char *endptr;
  int val = strtol(arg,&endptr,base);

  if (!(!*endptr && min <= val && val <= max)) {
    pout("File %s line %d : Directive: %s has argument: %s; needs "
         "integer from %d to %d.\n",
	   cfgfile, lineno, token, arg, min, max);
    return -1;
  }

  // all is well; return value
  return val;
}

const char * const parseopts = "ITVF";
static const char * parseptr = parseopts;
// This function returns 1 if it has correctly parsed one token (and
// any arguments)returns 2 if the all tokens of the threshold are parsed
// for a attrtube id("-I -T -V -F"), else it returns -1 if an error was
// encountered and -2 if the error msg is already logged.
static int ParseToken(char * token, int lineno, attr_thres_cfg & attr_cfg)
{

  char sym;
  const char * delim = " \n\t";
  const char * arg = 0;

  // is the rest of the line a comment
  if (*token=='#')
    return 1;

  // is the token not recognized?
  if (*token!='-' || strlen(token)!=2)
    return -1;

  int val;
  // token we will be parsing:
  sym=token[1];
  // check the token is the expected one
  if (*parseptr!=sym) {
    pout("File %s line %d : ERROR Directive -%c Expected Directive -%c\n",
	     configfile, lineno, sym, *parseptr);
    return -2;
  }

  switch (sym) {
    case 'I':
      // Attribute ID whose thershold to be monitored
      if ((val = GetValue(arg=strtok(NULL,delim), token,
			lineno, configfile, 0, 255, 10)) < 0)
	return -2;
      attr_cfg.attr_id = val;
      parseptr++;
      break;
    case 'T':
      // Thershold values for the Attribute ID specified by -I
      if ((val = GetValue(arg=strtok(NULL,delim), token,
			lineno, configfile, 0, 0xFFFFFF, 16)) < 0)
	return -2;
      attr_cfg.attr_thres = val;
      parseptr++;
      break;
    case 'V':
      // Which value between raw and normalised value should be compared
      // against threshold specified by -T
      attr_cfg.attr_flags = THRES_INVALID;
      arg = strtok(NULL, delim);
      if (!arg) {
	return -1;
      } else if (!strcmp("RAW", arg) ||
		 !strcmp("raw", arg)) {
	attr_cfg.attr_flags |= THRES_RAW;
      } else if (!strcmp("NORM", arg) ||
		 !strcmp("norm", arg)) {
	attr_cfg.attr_flags |= THRES_NORM;
      } else {
	return -1;
      }
      parseptr++;
      break;
    case 'F':
      // Error shall be reported when the value is above or below threshold
      arg = strtok(NULL, delim);
      if (!arg) {
       return -1;
      } else if (!strcmp("LOW", arg) ||
		 !strcmp("low", arg)) {
	attr_cfg.attr_flags |= THRES_LOW;
      } else if (!strcmp("HIGH", arg) ||
		 !strcmp("high", arg)) {
	attr_cfg.attr_flags |= THRES_HIGH;
      } else {
	return -1;
      }
      attr_cfg.attr_flags |= THRES_VALID;
      parseptr=parseopts;
      return 2;
      break;
    default:
      return -1;
  }
  return 1;
}

// This is the routine that adds things to the thres_db list.
//
// Return values are:
//  1: parsed a normal line and added a entry into thres_db.
//  0: found comment or blank line
// -1: found an error

// Note: this routine modifies *line from the caller!
static int ParseConfigLine(int lineno, char * line)
{
  char *token=NULL;
  char *name=NULL;
  const char *delim = " \n\t";

  // get first token: device name. If a comment, skip line
  if (!(name=strtok(line,delim)) || *name=='#') {
    return 0;
  }

  // The entires for threshold database
  attr_thres_cfg  attr_cfg;
  drive_thres_cfg thres_cfg(name);

  // parse tokens one at a time from the file.
  while ((token=strtok(NULL,delim))){
    int retval=ParseToken(token, lineno, attr_cfg);

    if (retval>0) {
      // Parsed token -- Enforcing Last token parsed -F (ret==2)
      // Need to modify here when enhancement is needed
      if (retval==2) {
        thres_cfg.push_back(attr_cfg);

        // Note - attr_flags includes THRES_VALID --
        // indicating this attr_cfg record is something !!!
        // Reset the attr_cfg for next record
        attr_cfg.attr_id = 0;
        attr_cfg.attr_thres = 0;
        attr_cfg.attr_flags = 0;
      }
      continue;
    }

    if (retval<0) {
      // error found on the line
      if (retval==-1)
        pout("File %s line %d : unknown Directive: %s\n."
             "Please refer to smartd_thres.conf for usage details.\n",
             configfile, lineno, token);
        return -1;
    }
  }
  thres_db.push_back(thres_cfg);
  return 1;
}

// Parses a configuration file.  Return values are:
//  N=>0: found N entries
// -1:    syntax error in config file
// -2:    config file does not exist
// -3:    config file exists but cannot be read
static int ParseThresConfigFile()
{
  // maximum line length in configuration file
  const int MAXLINELEN = 256;
  // maximum length of a continued line in configuration file
  const int MAXCONTLINE = 1023;

  int ret = -1;

  stdio_file f;

  if (!f.open(configfile,"r")) {
    // file exists but we can't read it
    ret = (errno!=ENOENT ? -3 : -2);
    pout("%s: Unable to open configuration file %s\n",
	     strerror(errno),configfile);
    return ret;
  }

  // configuration file exists
  pout("Opened configuration file %s\n",configfile);

  // parse config file line by line
  int lineno = 1, cont = 0, contlineno = 0;
  char line[MAXLINELEN+2];
  char fullline[MAXCONTLINE+1];

  for (;;) {
    int len=0;
    char *lastslash;
    char *comment;
    contlineno++;

    // make debugging simpler
    memset(line,0,sizeof(line));

    // get a line, if we are at the end of line break
    if (!(fgets(line, MAXLINELEN+2, f)))
      break;


    // See if line is too long
    len=strlen(line);
    if (len>MAXLINELEN){
      const char *warn;
      if (line[len-1]=='\n')
	warn="(including newline!) ";
      else
	warn="";
      pout("Error: line %d of file %s %sis more than MAXLINELEN=%d \
characters.\n",
	       (int)contlineno,configfile,warn,(int)MAXLINELEN);
      return -1;
    }

    // Ignore anything after comment symbol
    if ((comment=strchr(line,'#'))){
      *comment='\0';
      len=strlen(line);
    }

    // is the total line (made of all continuation lines) too long?
    if (cont+len>MAXCONTLINE){
      pout("Error: continued line %d (actual line %d) of file %s is \
more than MAXCONTLINE=%d characters.\n",
	       lineno, (int)contlineno, configfile, (int)MAXCONTLINE);
      return -1;
    }

    // copy string so far into fullline, and increment length
    strcpy(fullline+cont,line);
    cont+=len;

    // is this a continuation line.  If so, replace \ by space
    // and look at next line
    if ( (lastslash=strrchr(line,'\\')) && !strtok(lastslash+1," \n\t")){
      *(fullline+(cont-len)+(lastslash-line))=' ';
      continue;
    }

    ret = ParseConfigLine( contlineno, fullline);
    if (ret<0) {
      // Error parsing the file close the file and clear the threshold db list
      f.close();
      thres_db.clear();
      return -1;
    }

    lineno++;
    cont=0;

  }

#ifdef DEBUG_THRES_CONFIG_READ
  {
    attr_thres_cfg  att;
    for (int i=0; i< thres_db.size(); i++) {
      drive_thres_cfg db = thres_db[i];

      for (int k=0; k< db.size();  k++) {
	att = db[k];
        printf("%s Id:%d Thres:0x%08X 0x%04x \n",
		db.get_drive_name(), att.attr_id,
		att.attr_thres, att.attr_flags);
      }
    }
  }
#endif

  f.close();
  return thres_db.size();
}

// Get the path of the threshold configuartion file
static const char * get_thresconfigfile_path()
{
  struct stat st;

#ifndef _WIN32
  if (stat(SMARTMONTOOLS_SYSCONFDIR"/smartd_working.conf", &st) != 0) 
    return SMARTMONTOOLS_SYSCONFDIR"/smartd_thres.conf";
  else
    return SMARTMONTOOLS_SYSCONFDIR"/smartd_working.conf";
#else
  // static std::string configfile_str = get_exe_dir() + "/smartd_thres.conf";
  static std::string configfile_str = get_exe_dir() + "/smartd_working.conf";

  return configfile_str.c_str();
#endif
}

// This function parses the configuration file and
// builds the threshold database.
int ReadDriveThresConfig()
{
  //Empty the data base and  update new thresholds
  thres_db.clear();
  parseptr=parseopts;
  configfile = get_thresconfigfile_path();

  return ParseThresConfigFile();
}

// check if the tresh_db is not empty
bool IsThresDbExists()
{
  return !(thres_db.empty());
}

#define SMARTD_DEFAULT_DEVICE "DEFAULT_SSD"
// this function updates the threshold for the corresponding
// attribube ID it returns true if successfull and false if
// if the attribute ID is not a valid one
bool UpdateConfigThersholds(const char *device_name,
			    const char * model,
			    struct  ata_smart_values * smart_val,
			    struct ata_smart_threshold_cfg * attr_thres)
{
  // iterate over the thres_db list
  for (unsigned i=0; i<thres_db.size(); i++) {
    // Get the config for one drive
    drive_thres_cfg thres_cfg=thres_db.at(i);
    // The series name to be configured
    const char * series_name=thres_cfg.get_drive_name();
    /* the config file may contain partial series name */
    if ((strstr(model, series_name)==NULL) &&
	(strcmp(SMARTD_DEFAULT_DEVICE, series_name)))
      continue;
    // iterate over the threshold config for the device and update the
    // threshold values and flags
    for (unsigned j=0; j<thres_cfg.size(); j++) {
      // theshold config for each attribute configured
      attr_thres_cfg  attr_cfg=thres_cfg[j];
      unsigned char id=attr_cfg.attr_id;
      int idx=ata_find_attr_index(id, *smart_val);
      if (idx<0) {
        pout("Device %s:Error: The attribute ID: %d, "
	     "is not valid for drive series %s\n",
		device_name, attr_cfg.attr_id, series_name);
        return false;
      }

      /*
       * Custom made embedded threshold value for UNIGEN
       * Note- Support & limitation
       *   - Two lines / Records of ID (209)
       *   -I 209 -T 0x14 -V RAW -F LOW
       *   -I 209 -T 0x05 -V RAW -F LOW
       *   - LOW/HIGH both must be the same (not checking)
       *   - RAW/NORM both must be the same (not checking)
       * Logic-
       * for-loop hit the first 209 id record set the attr_thres
       * subsequent 209 record |= attr_thres << 16
       * becomes 0x00050014
       * save id as well. (added on top of linux patch)
       */
      if (attr_thres[idx].threshold) {
          attr_thres[idx].threshold |= attr_cfg.attr_thres << 16;
      }
      else {
          attr_thres[idx].threshold = attr_cfg.attr_thres;
      }
      attr_thres[idx].flags=attr_cfg.attr_flags;
      attr_thres[idx].id=attr_cfg.attr_id;
    }
    return true;
  }
  return false;
}

int GenericCheckSmart(const char *drive_name,
		      const ata_smart_values & curval,
		      struct ata_smart_threshold_cfg * attr_thres,
		      const ata_vendor_attr_defs & defs)
{
    uint64_t value=0;
    const char *delim = " \n\t";
    char d_name[200];
    int  i;
    int  ret = 0;

    for (int i = 0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {
        if (attr_thres[i].flags & THRES_VALID) {

            if (attr_thres[i].flags & THRES_RAW)
              value = ata_get_attr_raw_value(
			curval.vendor_attributes[i], defs);
            else if (attr_thres[i].flags & THRES_NORM)
              value = curval.vendor_attributes[i].current;

	    if (((attr_thres[i].flags & THRES_LOW) &&
                (value <= attr_thres[i].threshold)) ||
                ((attr_thres[i].flags & THRES_HIGH) &&
                (value >= attr_thres[i].threshold))) {
                std::string attrname = ata_get_smart_attr_name(
			curval.vendor_attributes[i].id, defs);
                pout("Device %s:Failed SMART usage Attribute: %d %s.\n",
			drive_name, curval.vendor_attributes[i].id, 
			attrname.c_str());

                return attr_thres[i].id;
            }
        }
    }
    return ret;
}

/* The function returns the raw value for unigen */
/* From old smartd */
static u_int64_t
getrawvalue(unsigned char *raw)
{
    int i;
    u_int64_t rawvalue = 0;

    for (i = 5; i >= 0 ; i--) {
        rawvalue = rawvalue << 8;
        rawvalue |= raw[i];
    }
    return rawvalue;
}

/* the function UnigenCheckSmart implements unigen specific
 * smart check algorithm.
 * Attribute ID 209 - Remaining Life (%) configured in smart.conf
 * a)   This attribute is normalized to 100 in the Raw Value field.
 * b)   This number decreases.
 * c)   Read Raw Value (in hex)
 * return value :
 * id(209) :  Case 1: If Attribute 195 + Attribute 196 >= 0x8.
 *                    If the Raw Value field decreases to 0x14 (20 decimal).
 *                    This indicates 20% remaining spare blocks.
 *            Case 2: If Attribute 195 + Attribute 196 < 0x8.
 *                    If the Raw Value field decreases to 0x5.
 *                    This is indicates 5% remaining life.
 * 0       : if raw value is with the limits mentioned above.
 */
int UnigenCheckSmart(const char *drive_name,
                      const ata_smart_values & curval,
                      struct ata_smart_threshold_cfg * attr_thres,
                      const ata_vendor_attr_defs & defs)
{
    int i, idx = -1, found = 0;
    struct ata_smart_attribute_unigen *data, *attr_to_check;
    u_int64_t value = 0, rawval_195 = 0, rawval_196 = 0;
    unsigned int thresl;

    /*
     * Loop through all attributes (0-30)
     * read raw-value 195/196 from smart-data-read-from-SSD
     * get the attr-idx to be testified on (idx_to_test)
     * It is the one with THRES_VALID marked in flags
     *
     * Perform Case1&2 test logic after the loop and
     * act on the result.
     */
    for (i=0; i < NUMBER_ATA_SMART_ATTRIBUTES; i++) {

	/* Note-
	 * The vendor_attributes structure for unigen
	 * is different from typical ata_smart_attribute
	 */
        data = (struct ata_smart_attribute_unigen*)
		&curval.vendor_attributes[i];

#ifdef DEBUG_THRES_CONFIG_READ
	if (data->id == 195 ||
	    data->id == 196 ||
	    (attr_thres[i].flags & THRES_VALID)) {
            printf("UnigenCheckSmart: Id:%d Thres:0x%08X 0x%04x \n",
		attr_thres[i].id, attr_thres[i].threshold,
		attr_thres[i].flags);
	}
#endif

	switch (data->id) {
	    case 195:
		/* Get Raw Value of attr 195 */
                rawval_195 = getrawvalue(data->raw);
		break;
	    case 196:
		/* Get Raw Value of attr 196 */
                rawval_196 = getrawvalue(data->raw);
		break;
	    default:
	        /*
	         * Limitation: support only one field defined
		 * in smartd_thres.conf
		 */
		if (attr_thres[i].flags & THRES_VALID) {
		    attr_to_check = data;
		    idx = i;
		    found = 1;
		}
	}

	/*
	 * Currently, test on attr 209 (which is > 195/196)
	 * If we are to test attr < 195/196, should not
	 * break and run through the entire for-loop-attr
	 */
	if (found) {
	    break;
	}
    }

    if (found && (idx >= 0)) {
	/* check on RAW or NORMalized? defined as -V RAW/NORM */
        if (attr_thres[idx].flags & THRES_RAW) {
            value = getrawvalue(attr_to_check->raw);
        } else {
            value = attr_to_check->normalized;
	}

	/*
	 * the two threshold limits for the id are stored in
         * threshold as low 16 bits for 1st threshold and
         * high 16bits for the 2nd thershold
         */
	if (rawval_195 + rawval_196 >= 0x08) {
            thresl = (attr_thres[idx].threshold) & 0xFFFF;
        } else {
            thresl = ((attr_thres[idx].threshold) >> 16) & 0xFFFF;
        }

	/* check against high or low threshold limits */
        if(attr_thres[idx].flags & THRES_LOW) {
            if (value <= thresl) {
                std::string attrname =
			ata_get_smart_attr_name(attr_to_check->id, defs);
                pout("Device %s:Failed SMART usage Attribute: %d %s.\n",
			drive_name, attr_to_check->id, attrname.c_str());

		// Test Failed - Return thres-ID indicates its failure
                return attr_thres[idx].id;
            }
        } else {
            // It is THRES_HIGH
            if (value >= thresl) {
                std::string attrname =
			ata_get_smart_attr_name(attr_to_check->id, defs);
                pout("Device %s:Failed SMART usage Attribute: %d %s.\n",
			drive_name, attr_to_check->id, attrname.c_str());

		// Test Failed - Return thres-ID indicates its failure
                return attr_thres[idx].id;
            }
        }
    }

  // Test Succeeded - Return 0
  return 0;
}

// The followings are added from old smartd (945442) for unigen attribute names
#define ATTR_SIZE 30
static char vendor_attribute_name[ATTR_SIZE];

char *
atagetunigenattribname(unsigned char id)
{
    int i;
    for (i = 0; ((i < NUMBER_ATA_SMART_ATTRIBUTES) &&
                (unigen_attr[i].id != 0)); i++) {
        if ( unigen_attr[i].id == id) {
            snprintf(vendor_attribute_name, ATTR_SIZE, "(%3d)%s ", id,
                     unigen_attr[i].vendor_attrib_name);
            return vendor_attribute_name ;
        }
    }
    if (id >= 201) {
        snprintf(vendor_attribute_name, ATTR_SIZE,
                 "(%3d)Vendor Unique           ", id);
    } else {
        snprintf(vendor_attribute_name, ATTR_SIZE,
                 "(%3d)Unknown Attribute       ", id);
    }
    return vendor_attribute_name ;
}
